'' System Version 2.002
'' OS_loader Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/10/25 - Added bootAddr definition
''   2006/11/03 - Changed loader methods to use low speed
''   2006/11/04 - Added writeWait method to wait for write completion
''                Added support for non-OS programs to bootEEPROM
''   2006/11/05 - Added altBootAddr for relocated boot EEPROM
''   2006/11/06 - Limited boot loading to 32K - 16 (OS uses last 16 bytes)
''                Added verifyEEPROM method
''   2006/11/12 - Modified for new VGA driver
''   2006/11/14 - Hardware dependent information moved to this object
''                bootEEPROM now stops unregistered cogs if using OS conventions
''   2006/12/08 - Cleaned up definitions related to display

'' This portion is intended to be incorporated into programs and is the loader/I2C
'' application interface.

'' This object provides an I2C EEPROM read/write driver set up for 100KHz or 400KHz bus
'' speeds and sequential reads and paged writes.  A special read mode will stop the caller's
'' COG before overwriting the current program, then restart the SPIN interpreter with the
'' new program.  The control information is passed in a 2 long parameter block whose address
'' is passed to the read/write driver when it is started.  The parameter block is updated
'' when the operation completes with the current device address, byte count, and HUB address.
'' An error code is posted in the command/status byte.  Note that the following layout is
'' based on long values, not on the byte layout in memory.  The bus speed and page size
'' can be configured at compilation and can not currently be changed at run time.

'' -------------------------------------------------------------------
'' |   cmd/status   |          I/O pin / device / address            |
'' -------------------------------------------------------------------
'' |           byte count           |          HUB address           |
'' -------------------------------------------------------------------

'' The EEPROM address is in the same format used by other routines with the I/O pin pair
'' in bits 21..19, the device address in bits 18..16, and the 64K address in bits 15..0.
'' Note that the I/O pin pair is the number of the SCL pin divided by 2.  The SDA pin is
'' always the next higher numbered pin.  The command code is in the low order bits of the
'' high order byte of the first long (see ioCmdMask).  This is always non-zero to indicate
'' that a command is to be performed by the COG routines.  When the command is finished,
'' this is set to zero.  The errorFlag bit is set to one if a NAK was read after a write
'' transfer.  This is the only error reported by these routines.  A read operation and
'' zero-length writes do involve several write transfers for addressing, but the data
'' read transfer has no error checking.  When the command is completed, the device address,
'' byte count, and HUB address are all updated to their values at that time.

'' Command codes are provided for devices with zero, one, or two address bytes to follow
'' the device selection byte.  As for all I2C devices, addressing is done using write
'' mode and the device is reselected in read mode after the last address byte.  In the
'' case of ioRead0Cmd, the device is initially selected in read mode.  For 8-bit addresses,
'' the device select code is taken from bits 15-8 of the address value.  For the case
'' without address bytes, the device select code is taken from bits 7-0 of the address value.

'' The pins used for the boot EEPROM I2C bus (at least on Parallax's Demo Board) do not
'' have a pullup on SCL.  This requires that SCL be driven both high and low.  If the bus
'' used is on pins 28 and 29, SCL is actively driven at all times.

'' This object also provides constant declarations for sizes of various system tables and
'' buffers and offsets to their pointers in "high memory".  The offsets within the system
'' work area are declared here as well.  Some simple methods are provided for memory allocation
'' and for registering and unregistering COGs as being persistant.  Any COGs not registered
'' will be stopped when a program is read into RAM using the RUN command.  These routines
'' depend on the global area being initialized by the "Operating System" initialization
'' routines and these should not be called unless the global area has been initialized since
'' the last reset.  The EEPROM read, write, boot, and "check presence" routines are not
'' dependent on this initialization (although they do use some high memory locations) and
'' can be used without the rest of the "Operating System" routines.

CON
  OSversion   = 2
  OSrelease   = 002
  OStestLevel = 0

'' Hardware dependent and other basic definitions

  TV          = false                          ' Display type
  VGA         = not TV

  Parallax    = true                           ' Parallax Demo Board conventions
                                               ' Keyboard on pins 26-27, 2 pins/device
                                               ' (the same pins used for input & output)
                                               ' TV on pins 8-11, VGA on pins 16-23
                                               ' 5 MHz crystal with PLL x 16
  Hydra       = not Parallax                   ' Hydra Game System conventions
                                               ' Keyboard on pins 12-15, 4 pins/device
                                               ' (pairs of pins used for input & output)
                                               ' TV on pins 24-27, VGA on pins 16-23
                                               ' 10 MHz crystal with PLL x 8

  cursor      = false                          ' Cursor function enabled
  tiles       = (TV & 1) | (VGA & 2)           ' Tiles per character (1-16x32, 2-16x16)
                                               ' TV must be 1, VGA looks better with 2
  colors      = 64                             ' Different possible colors (2^n)
  windows     = 32                             ' Number of windows supported (2^n)
  device      = (TV & 0) | (VGA & 1)           ' Device type
  cols        = (TV & 40) | (VGA & 64)         ' Display character columns
  rows        = (TV & 13) | (VGA & 24)         ' Display character rows

  keyPinGroup = (Parallax & 26) | (Hydra & 12)
  tvPinGroup  = (Parallax & 12) | (Hydra & 24)
  dspPinGroup = (TV & tvPinGroup) | (VGA & 16)
  clockMode   = xtal1 | (Parallax & pll16x) | (Hydra & pll8x)
  xtalClock   = (Parallax & 5_000_000) | (Hydra & 10_000_000)

'' General hardware dependent definitions

  maxPinPairs = 16                             ' Maximum number of possible I2C busses
  i2cBootSCL  = 28                             ' I2C Clock pin for boot EEPROM
  i2cBootSDA  = 29                             ' I2C Data pin for boot EEPROM
  bootAddr    = i2cBootSCL << 18               ' File address for boot EEPROM
  altBootAddr = bootAddr + 128*1024            ' Relocated boot EEPROM address (for Hydra)

'' System work area definitions

  sys_cog     = 0                              ' Contains bitmap for cogs used by system
  sys_cogLdr  = 2                              ' Contains cog # + 1 for I2C I/O driver
  sys_cogMain = 3                              ' Contains cog # + 1 for main SPIN task
  sys_random  = 4                              ' Random seed for "?" operator
  sys_eeproms = 8                              ' Table of EEPROMs, one for each I/O pair
                                               ' Each word entry contains a bitmap of the
                                               ' 32Kx8 address blocks that are present on
                                               ' the I2C bus on the corresponding pair of
                                               ' I/O pins (if any).

'' System work areas, buffer, and table sizes

'' ***** IMPORTANT NOTE *****

'' These table sizes are copied from the various other modules in this OS to avoid
'' circular object references.  The sizes of the video tables (videoSize, screenSize,
'' colorSize, cursorSize, and fontSize) will vary depending on which driver is used.
'' If there are any modification of the other modules, make sure these are updated.

  systemSize  = sys_eeproms + maxPinPairs*2    ' # of bytes for system information
  tvBase      = 84                             ' # of bytes for TV information
  vgaBase     = 32                             ' # of bytes for VGA information
  videoSize   = ((TV & tvBase) | (VGA & vgaBase)) + windows * 8
  screenSize  = cols * rows * tiles * 2        ' # of bytes for screen buffer
  colorSize   = colors * 4                     ' # of bytes for color buffer
  cursorSize  = cursor & (4 * 4)               ' # of bytes for screen cursor info
  fontSize    = 0                              ' # of bytes for font table
  keybdSize   = 80                             ' # of bytes needed for keyboard
  videoNeeded = videoSize + screenSize + colorSize + cursorSize + fontSize
  totalNeeded = systemSize + videoNeeded + keybdSize

'' The following are fixed locations assigned at the top of RAM memory to avoid
'' having them overwritten when a new program is loaded.  Most of these are pointers
'' to work areas allocated "below" them.  The last pointer is reserved for non-system
'' programs

  endMemory   = $8000                          ' Address past the end of RAM
  initMarker  = endMemory -  4                 ' Unique marker for initialized data
  memPtr      = endMemory -  8                 ' Address of last allocated long
  systemPtr   = endMemory - 12                 ' Address of system information
  videoPtr    = endMemory - 16                 ' Address of video work area
  screenPtr   = endMemory - 20                 ' Address of screen buffer
  colorPtr    = endMemory - 24                 ' Address of color buffer
  cursorPtr   = endMemory - 28                 ' Address of screen cursor info
  fontPtr     = endMemory - 32                 ' Address of font table for screen
  keybdPtr    = endMemory - 36                 ' Address of keyboard work area
  userPtr     = endMemory - 40                 ' Address of work area for "user"
  ioControl   = endMemory - 48                 ' I/O control block for general use
  endFree     = ioControl                      ' Address past initial free memory

  uniqueMark  = $965A3EC1                      ' Indicates that RAM was initialized
  noSuchAddr  = $8000                          ' "Invalid" address for initializing pointers
  
'' Internal SPIN information at fixed locations in RAM/EEPROM image

  spin_pbase  = $0006                          ' must be $0010
  spin_vbase  = $0008                          ' number of longs loaded times 4
  spin_dbase  = $000A                          ' above where $FFF9FFFF's get placed
  spin_pcurr  = $000C                          ' points to SPIN code
  spin_dcurr  = $000E                          ' points to local stack

'' Command code and error information for I2C driver

  ioReadCmd   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  ioWriteCmd  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  ioRead1Cmd  = %00000011                      ' Read from a device with only 8-bit addresses
  ioWrite1Cmd = %00000100                      ' Write to a device with only 8-bit addresses
  ioRead0Cmd  = %00000101                      ' Read from a device without address bytes
  ioWrite0Cmd = %00000110                      ' Write to a device without address bytes
  ioVerifyCmd = %00000111                      ' Like ioReadCmd, but doesn't store to HUB RAM
                                               ' Data is checksummed and an error is reported
                                               '  if the checksum is not zero
  ioBootCmd   = %00001000                      ' Read from EEPROM to HUB RAM, then start a
                                               ' new SPIN interpreter in the COG whose ID is
                                               ' supplied in the lower 3 bits of this command
                                               ' This COG is stopped before the read is done.
  ioCmdMask   = %00001111                      ' Used to mask off command bits
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

'' Option flags information from OS_fileSystem (Keep this copy updated if changed)

  opt_text    = %0000_0000_0001                ' File is text (not binary)
  opt_noOS    = %0000_0000_0010                ' Disable OS while loading program

PUB bootEEPROM(addr) | t, s, p                         '' Boot SPIN from a block in EEPROM
  p := long[systemPtr]
  if addr & opt_noOS                                   ' If the program doesn't use this OS,
    s := byte[p+sys_cogLdr] - 1                        '  stop all COGs except this one and
    repeat t from 0 to 7                               '   the one with the I2C driver in it
      if (t <> cogid) and (t <> s)
        cogstop(t)                                     ' Mark the OS as no longer running
    long[initMarker] := 0                              '  and use the full 32K-16 for loading
    s := $7FF0                                         '   or the actual program size
  else
    repeat t from 0 to 7                               ' Stop any cogs not registered and
      if (word[p+sys_cog] & |<t) == 0                  '  use memory up to the high memory
        cogstop(t)                                     '   areas used by the OS
    s := long[memPtr]
  t := ioBootCmd | ioLowSpeed | cogid                  ' Prepare to stop COG of caller
  repeat while long[ioControl] & ioTestRdy             ' Wait for previous I/O to finish
  long[ioControl+4] := (s << 16) | 0
  long[ioControl] := (t << 24) | (addr & $FFF000)      ' Mask off option bits
  repeat while long[ioControl] & ioTestRdy             ' Wait for this to finish if errors
  return (long[ioControl] & ioTestErr) == 0            ' Return false if error occurred

PUB readEEPROM(addr,buffer,count) | t                  '' Read a block from EEPROM to RAM
  t := ioReadCmd | ioLowSpeed
  repeat while long[ioControl] & ioTestRdy             ' Wait for previous I/O to finish
  long[ioControl+4] := (count << 16) | (buffer & $FFFF)
  long[ioControl] := (t << 24) | (addr & $FFFFFF)
  repeat while long[ioControl] & ioTestRdy             ' Wait for this to finish
  return (long[ioControl] & ioTestErr) == 0            ' Return true if successful

PUB verifyEEPROM(addr) | t, s                          '' Verify (checksum) a block of EEPROM
  if addr & opt_noOS                                   ' If the program doesn't use this OS,
    s := $7FF0                                         '  use the full 32K-16 for verifying
  else                                                 '   or the actual program size
    s := long[memPtr]
  t := ioVerifyCmd | ioLowSpeed                        ' Report an error if checksum not zero
  repeat while long[ioControl] & ioTestRdy             ' Wait for previous I/O to finish
  long[ioControl+4] := (s << 16) | 0
  long[ioControl] := (t << 24) | (addr & $FFF000)      ' Mask off option bits
  repeat while long[ioControl] & ioTestRdy             ' Wait for this to finish
  return (long[ioControl] & ioTestErr) == 0            ' Return true if successful

PUB writeEEPROM(addr,buffer,count) | t                 '' Write a block to EEPROM from RAM
  t := ioWriteCmd | ioLowSpeed
  repeat while long[ioControl] & ioTestRdy             ' Wait for previous I/O to finish
  long[ioControl+4] := (count << 16) | (buffer & $FFFF)
  long[ioControl] := (t << 24) | (addr & $FFFFFF)
  repeat while long[ioControl] & ioTestRdy             ' Wait for this to finish
  return (long[ioControl] & ioTestErr) == 0            ' Return true if successful

PUB checkPresence(addr) | t
'' This routine checks to be sure there is an I2C bus and an EEPROM at the
'' specified address.  Note that this routine cannot distinguish between a
'' 32Kx8 and a 64Kx8 EEPROM since the 16th address bit is a "don't care"
'' for the 32Kx8 devices.  Return true if successful, false otherwise.
  t := ioReadCmd | ioLowSpeed
  repeat while long[ioControl] & ioTestRdy             ' Wait for previous I/O to finish
  long[ioControl+4] := 0                               ' Attempt to address the device
  long[ioControl] := (t << 24) | (addr & $FFFFFF)
  repeat while long[ioControl] & ioTestRdy             ' Wait for this to finish
  return (long[ioControl] & ioTestErr) == 0            ' Return true if successful

PUB writeWait(addr) | t                                '' Wait for EEPROM to complete write
  t := cnt                                             ' Maximum wait time is 20ms
  repeat until checkPresence(addr)
    if (cnt - t) > (clkfreq / 50)
      return false                                     ' Return false if a timeout occurred
  return true                                          ' Otherwise return true

PUB allocatePRI(size)                                  '' Allocate block of size bytes
  return long[memPtr] := (long[memPtr] - size) & !3    ' Always a long address

PUB allocate(size)                                     '' Allocate a block of user work area
  if long[userPtr] <> noSuchAddr
    deallocate                                         ' Deallocate current area if allocated
  return long[userPtr] := allocatePRI(size)

PUB deallocate                                         '' User work area no longer in use
  long[memPtr] := long[keybdPtr]
  long[userPtr] := noSuchAddr                          ' Use ROM address for "invalid" address

PUB registerCOG(bitmap)                                '' Register "persistant" COGs
  word[long[systemPtr]+sys_cog] |= bitmap

PUB unregisterCOG(bitmap)                              '' Mark COGs no longer "persistant"
  word[long[systemPtr]+sys_cog] &= !bitmap

PUB start(addr) | p, c                                 '' Start I2C I/O driver given its address
  p := long[systemPtr]                                 ' Note: computeTimes must be called first!!
  stop                                                 ' Stop read/write COG if running
  long[ioControl] := 0                                 ' Now clear the I/O control parameters
  long[ioControl+4] := 0                               '  including byte count & hub address
  c := cognew(addr,ioControl) + 1                      ' Start a fresh copy
  if c
    byte[p+sys_cogLdr] := c                            ' If successful, register the COG
    registerCOG(|<(c - 1))
  return c > 0                                         ' Return success

PUB stop | p, c                                        '' Stop a running I2C I/O driver
  p := long[systemPtr]
  c := byte[p+sys_cogLdr]                              ' Get the saved cog # (+ 1)
  if c
    cogstop(c - 1)                                     ' If running, stop it
    unregisterCOG(|<(c - 1))                           '  and unregister it
  byte[p+sys_cogLdr] := 0